__all__ = ('EmbedFooter',)

import warnings

from scarletio import copy_docs

from ...utils import sanitize_mentions, url_cutter

from ..embed_field_base import EmbedFieldBase

from .fields import (
    parse_icon_proxy_url, parse_icon_url, parse_text, put_icon_proxy_url_into, put_icon_url_into, put_text_into,
    validate_icon_url, validate_text
)


class EmbedFooter(EmbedFieldBase):
    """
    Represents an embed's footer.
    
    Attributes
    ----------
    icon_url :`None`, `str`
        Url of the embed footer's icon.
    icon_proxy_url : `None`, `str`
        A proxied url of the embed footer's icon.
    text : `None`, `str`
        The embed footer's text.
    """
    __slots__ = ('icon_url', 'icon_proxy_url', 'text')
    
    def __new__(cls, text = ..., icon_url = ...):
        """
        Creates a new embed field with the given parameters.
        
        Parameters
        ----------
        text : `None`, `str`, Optional
            The footer's text.
        icon_url : `None`, `str`, Optional
            An url of the footer's icon. Can be http(s) or attachment.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # icon_url
        if icon_url is ...:
            icon_url = None
        else:
            icon_url = validate_icon_url(icon_url)
        
        # text
        if text is ...:
            text = None
        else:
            text = validate_text(text)
        
        self = object.__new__(cls)
        self.icon_url = icon_url
        self.icon_proxy_url = None
        self.text = text
        return self
    
    
    @copy_docs(EmbedFieldBase.__len__)
    def __len__(self):
        length = 0
        
        # text
        text = self.text
        if (text is not None):
            length += len(text)
        
        return length
    
    
    @copy_docs(EmbedFieldBase.__bool__)
    def __bool__(self):
        # icon_url
        if (self.icon_url is not None):
            return True
        
        # text
        if self.text is not None:
            return True
        
        return False
    
    
    @copy_docs(EmbedFieldBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        field_added = False
        
        text = self.text
        if text is not None:
            field_added = True
            repr_parts.append(' text = ')
            repr_parts.append(repr(self.text))
        
        icon_url = self.icon_url
        if icon_url is not None:
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' icon_url = ')
            repr_parts.append(repr(url_cutter(icon_url)))
    
    
    @copy_docs(EmbedFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # icon_url
        icon_url = self.icon_url
        if (icon_url is not None):
            hash_value ^= hash(icon_url)
        
        # text
        text = self.text
        if (text is not None):
            hash_value ^= hash(text)
        
        return hash_value
    
    
    @copy_docs(EmbedFieldBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # icon_url
        if self.icon_url != other.icon_url:
            return False
        
        # text
        if self.text != other.text:
            return False
        

        return True
    
    
    @classmethod
    @copy_docs(EmbedFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.icon_url = parse_icon_url(data)
        self.icon_proxy_url = parse_icon_proxy_url(data)
        self.text = parse_text(data)
        return self
    
    
    @copy_docs(EmbedFieldBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_icon_url_into(self.icon_url, data, defaults)
        put_text_into(self.text, data, defaults)
        
        if include_internals:
            put_icon_proxy_url_into(self.icon_proxy_url, data, defaults)
        
        return data
    
    
    @copy_docs(EmbedFieldBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        new.icon_url = self.icon_url
        new.icon_proxy_url = self.icon_proxy_url
        new.text = sanitize_mentions(self.text, guild)
        return new
    
    
    @copy_docs(EmbedFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.icon_url = self.icon_url
        new.icon_proxy_url = self.icon_proxy_url
        new.text = self.text
        return new
    
    
    def copy_with(self, *, icon_url = ..., text = ...):
        """
        Copies the embed footer with the given parameters.
        
        Parameters
        ----------
        icon_url : `None`, `str`, Optional (Keyword only)
            An url of the footer's icon. Can be http(s) or attachment.
        text : `None`, `str`, Optional (Keyword only)
            The footer's text.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # icon_url
        if icon_url is ...:
            icon_url = self.icon_url
            icon_proxy_url = self.icon_proxy_url
        else:
            icon_url = validate_icon_url(icon_url)
            icon_proxy_url = None
        
        # text
        if text is ...:
            text = self.text
        else:
            text = validate_text(text)
        
        new = object.__new__(type(self))
        new.icon_url = icon_url
        new.icon_proxy_url = icon_proxy_url
        new.text = text
        return new
    
    
    @copy_docs(EmbedFieldBase.iter_contents)
    def iter_contents(self):
        text = self.text
        if (text is not None):
            yield text
    
    
    @property
    def proxy_icon_url(self):
        """
        Deprecated and will be removed in 2023 august. Please use ``.icon_proxy_url``.
        """
        warnings.warn(
            (
                f'`{type(self).__text__}.proxy_icon_url` is deprecated and will be removed in 2023 august. '
                f'Please use `.icon_proxy_url` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
    
        return self.icon_proxy_url



__all__ = ()

from ...field_parsers import (
    entity_id_parser_factory, flag_parser_factory, force_string_parser_factory, nullable_entity_parser_factory,
    nullable_functional_parser_factory, nullable_string_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, flag_optional_putter_factory, force_string_putter_factory,
    nullable_entity_optional_putter_factory, nullable_functional_optional_putter_factory,
    nullable_string_optional_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    nullable_date_time_validator_factory, nullable_entity_validator_factory, nullable_string_validator_factory,
    url_optional_validator_factory
)
from ...utils import datetime_to_millisecond_unix_time, millisecond_unix_time_to_datetime

from ..activity_assets import ActivityAssets
from ..activity_party import ActivityParty
from ..activity_secrets import ActivitySecrets
from ..activity_timestamps import ActivityTimestamps

from .flags import ActivityFlag

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id_into = entity_id_optional_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', NotImplemented, include = 'Application')

# assets

parse_assets = nullable_entity_parser_factory('assets', ActivityAssets)
put_assets_into = nullable_entity_optional_putter_factory('assets', ActivityAssets)
validate_assets = nullable_entity_validator_factory('assets', ActivityAssets)

# created_at

def parse_created_at(data):
    """
    Parses when the activity was created from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Activity timestamp data.
    
    Returns
    -------
    created_at : `None`, `datetime`
    """
    created_at = data.get('created_at', None)
    if (created_at is not None):
        return millisecond_unix_time_to_datetime(created_at)


def put_created_at_into(created_at, data, defaults):
    """
    Puts the activity created at timestamp into the given data.
    
    Parameters
    ----------
    created_at : `None`, `datetime`
        Activity timestamps created_at.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (created_at is not None):
        if (created_at is not None):
            created_at = datetime_to_millisecond_unix_time(created_at)
        
        data['created_at'] = created_at
    
    return data


validate_created_at = nullable_date_time_validator_factory('created_at')

# details

parse_details = nullable_string_parser_factory('details')
put_details_into = nullable_string_optional_putter_factory('details')
validate_details = nullable_string_validator_factory('details', 0, 1024)

# emoji

parse_emoji = nullable_functional_parser_factory('emoji', NotImplemented, include = 'create_partial_emoji_from_data')
put_emoji_into = nullable_functional_optional_putter_factory(
    'emoji', NotImplemented, include = 'create_partial_emoji_data'
)
validate_emoji = nullable_entity_validator_factory('emoji', NotImplemented, include = 'Emoji')

# flags

parse_flags = flag_parser_factory('flags', ActivityFlag)
put_flags_into = flag_optional_putter_factory('flags', ActivityFlag())
validate_flags = flag_validator_factory('flags', ActivityFlag)

# id


def parse_id(data):
    """
    Parses out activity identifier from the given data
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Activity data.
    
    Returns
    -------
    activity_id : `int`
    """
    activity_id = data.get('id', None)
    if (activity_id is None):
        return 0
    
    try:
        return int(activity_id, base = 16)
    except ValueError:
        # Some activity types have `custom` id-s, at those cases we default back to `0`.
        return 0


def put_id_into(activity_id, data, defaults):
    """
    Puts the given activity identifier into a json serializable dictionary.
    
    Parameters
    ----------
    activity_id : `int`
        Activity's identifier.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if activity_id:
        data['id'] = format(activity_id, 'x')
    
    return data


validate_id = entity_id_validator_factory('activity_id')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, 1024)

# party

parse_party = nullable_entity_parser_factory('party', ActivityParty)
put_party_into = nullable_entity_optional_putter_factory('party', ActivityParty)
validate_party = nullable_entity_validator_factory('party', ActivityParty)

# secrets

parse_secrets = nullable_entity_parser_factory('secrets', ActivitySecrets)
put_secrets_into = nullable_entity_optional_putter_factory('secrets', ActivitySecrets)
validate_secrets = nullable_entity_validator_factory('secrets', ActivitySecrets)

# session_id

parse_session_id = nullable_string_parser_factory('session_id')
put_session_id_into = nullable_string_optional_putter_factory('session_id')
validate_session_id = nullable_string_validator_factory('session_id', 0, 1024)

# state

parse_state = nullable_string_parser_factory('state')
put_state_into = nullable_string_optional_putter_factory('state')
validate_state = nullable_string_validator_factory('state', 0, 1024)

# sync_id

parse_sync_id = nullable_string_parser_factory('sync_id')
put_sync_id_into = nullable_string_optional_putter_factory('sync_id')
validate_sync_id = nullable_string_validator_factory('sync_id', 0, 1024)

# timestamps

parse_timestamps = nullable_entity_parser_factory('timestamps', ActivityTimestamps)
put_timestamps_into = nullable_entity_optional_putter_factory('timestamps', ActivityTimestamps)
validate_timestamps = nullable_entity_validator_factory('timestamps', ActivityTimestamps)

# url

parse_url = nullable_string_parser_factory('url')
put_url_into = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url')



__all__ = ()

BOT_PUBLIC_DEFAULT = False
BOT_REQUIRE_CODE_GRANT_DEFAULT = False
OVERLAY_DEFAULT = False
OVERLAY_COMPATIBILITY_HOOK_DEFAULT = False
HOOK_DEFAULT = False


MAX_PARTICIPANTS_DEFAULT = 0

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 128

DESCRIPTION_LENGTH_MAX = 400



__all__ = ('InteractionMetadataApplicationCommand',)

from scarletio import copy_docs

from .application_command_autocomplete import InteractionMetadataApplicationCommandAutocomplete
from .fields import (
    parse_resolved, parse_target_id, put_resolved_into, put_target_id_into, validate_resolved, validate_target_id
)


class InteractionMetadataApplicationCommand(InteractionMetadataApplicationCommandAutocomplete):
    """
    Interaction metadata used when the interaction was triggered by an application command.
    
    Parameters
    ----------
    id : `int`
        The represented application command's identifier number.
    
    name : `str`
        The represented application command's name.
    
    options : `None`, `tuple` of ``InteractionOption``
        Application command option representations. Like sub-command or parameter.
    
    resolved : `None`, ``Resolved``
        Contains the received entities.
    
    target_id : `int`
        The interaction's target's identifier. Applicable for context commands.
    """
    __slots__ = ('resolved', 'target_id')
    
    @classmethod
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._create_empty)
    def _create_empty(cls):
        self = super(InteractionMetadataApplicationCommand, cls)._create_empty()
        self.resolved = None
        self.target_id = 0
        return self
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.copy)
    def copy(self):
        new = InteractionMetadataApplicationCommandAutocomplete.copy(self)
        
        resolved = self.resolved
        if (resolved is not None):
            resolved = resolved.copy()
        new.resolved = resolved
        
        new.target_id = self.target_id 
        
        return new
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        # resolved
        try:
            resolved = keyword_parameters.pop('resolved')
        except KeyError:
            pass
        else:
            self.resolved = validate_resolved(resolved)
        
        # target_id
        try:
            target_id = keyword_parameters.pop('target_id')
        except KeyError:
            pass
        else:
            self.target_id = validate_target_id(target_id)
        
        
        InteractionMetadataApplicationCommandAutocomplete._set_attributes_from_keyword_parameters(
            self, keyword_parameters
        )
    
    
    @classmethod
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.from_data)
    def from_data(cls, data, interaction_event):
        self = super(InteractionMetadataApplicationCommand, cls).from_data(data, interaction_event)
        self.resolved = parse_resolved(data, interaction_event)
        self.target_id = parse_target_id(data)
        return self
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.to_data)
    def to_data(self, *, defaults = False, interaction_event = None):
        data = InteractionMetadataApplicationCommandAutocomplete.to_data(
            self, defaults = defaults, interaction_event = interaction_event
        )
        put_resolved_into(self.resolved, data, defaults, interaction_event = interaction_event)
        put_target_id_into(self.target_id, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        field_added = InteractionMetadataApplicationCommandAutocomplete._put_attribute_representations_into(
            self, repr_parts
        )
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' resolved = ')
            repr_parts.append(repr(resolved))
        
        # target_id
        target_id = self.target_id
        if target_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' target_id = ')
            repr_parts.append(repr(target_id))
        
        return field_added
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete.__hash__)
    def __hash__(self):
        hash_value = InteractionMetadataApplicationCommandAutocomplete.__hash__(self)
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            hash_value ^= hash(resolved)
        
        # target_id
        hash_value ^= self.target_id
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataApplicationCommandAutocomplete._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not InteractionMetadataApplicationCommandAutocomplete._is_equal_same_type(self, other):
            return False
        
        # resolved
        if self.resolved != other.resolved:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        return True



__all__ = ()

from ...channel import Channel
from ...role import Role
from ...user import ClientUserBase

from .preinstanced import ApplicationCommandPermissionOverwriteTargetType

APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER = ApplicationCommandPermissionOverwriteTargetType.user
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE = ApplicationCommandPermissionOverwriteTargetType.role
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL = ApplicationCommandPermissionOverwriteTargetType.channel


def validate_application_command_permission_overwrite_target(target):
    """
    Validates the given application command's input target.
    
    Parameters
    ----------
    target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
            ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`, \
            ``ApplicationCommandPermissionOverwriteTargetType``, `int`)), `int`)
        The target entity of the application command permission overwrite.
    
    Returns
    -------
    target_type : ``ApplicationCommandPermissionOverwriteTargetType``
        The target entity's type.
    
    target_id : `int`
        The represented entity's identifier.
    
    Raises
    ------
    TypeError
        - If `target` was not given as any of the expected types & values.
    """
    # GOTO
    while True:
        if isinstance(target, Role):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, ClientUserBase):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, Channel):
            target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
            target_id = target.id
            target_lookup_failed = False
            break
        
        if isinstance(target, tuple) and len(target) == 2:
            target_type_maybe, target_id_maybe = target
            
            if isinstance(target_type_maybe, type):
                if issubclass(target_type_maybe, Role):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                elif issubclass(target_type_maybe, ClientUserBase):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                elif issubclass(target_type_maybe, Channel):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
                else:
                    target_lookup_failed = True
                    break
            
            elif isinstance(target_type_maybe, str):
                if target_type_maybe in ('Role', 'role'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                elif target_type_maybe in ('User', 'user'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                elif target_type_maybe in ('Channel', 'channel'):
                    target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL
                else:
                    target_lookup_failed = True
                    break
            
            elif isinstance(target_type_maybe, ApplicationCommandPermissionOverwriteTargetType):
                target_type = target_type_maybe
            
            elif isinstance(target_type_maybe, ApplicationCommandPermissionOverwriteTargetType.VALUE_TYPE):
                target_type = ApplicationCommandPermissionOverwriteTargetType.get(target_type_maybe)
            
            else:
                target_lookup_failed = True
                break
            
            if type(target_id_maybe) is int:
                target_id = target_id_maybe
            elif isinstance(target_id_maybe, int):
                target_id = int(target_id_maybe)
            else:
                target_lookup_failed = True
                break
            
            target_lookup_failed = False
            break
        
        target_lookup_failed = True
        break
    
    if target_lookup_failed:
        raise TypeError(
            f'`target` can be `{Role.__name__}`, `{ClientUserBase.__name__}`, `{Channel.__name__}`, '
            f'`tuple` ((`{Role.__name__}`, `{ClientUserBase.__name__}`, `{Channel.__name__}`, `str` '
            f'(`\'Role\'`, `\'role\'`, `\'User\'`, `\'user\'`, `\'Channel\'`, `\'channel\'`)), `int`), '
            f'got {target.__class__.__name__}: {target!r}.'
        )
    
    return target_type, target_id



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, nullable_date_time_parser_factory, nullable_sorted_array_parser_factory,
    preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_string_array_optional_putter_factory, preinstanced_array_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, nullable_date_time_validator_factory, nullable_string_array_validator_factory,
    preinstanced_array_validator_factory, preinstanced_validator_factory
)

from ..discovery_category import DiscoveryCategory

# application_actioned

parse_application_actioned = nullable_date_time_parser_factory('partner_actioned_timestamp')
put_application_actioned_into = nullable_date_time_optional_putter_factory('partner_actioned_timestamp')
validate_application_actioned = nullable_date_time_validator_factory('application_actioned')

# application_requested

parse_application_requested = nullable_date_time_parser_factory('partner_application_timestamp')
put_application_requested_into = nullable_date_time_optional_putter_factory('partner_application_timestamp')
validate_application_requested = nullable_date_time_validator_factory('application_requested')

# emoji_discovery

parse_emoji_discovery = bool_parser_factory('emoji_discoverability_enabled', False)
put_emoji_discovery_into = bool_optional_putter_factory('emoji_discoverability_enabled', False)
validate_emoji_discovery = bool_validator_factory('emoji_discovery', False)

# keywords

parse_keywords = nullable_sorted_array_parser_factory('keywords')
put_keywords_into = nullable_string_array_optional_putter_factory('keywords')
validate_keywords = nullable_string_array_validator_factory('keywords')

# primary_category

parse_primary_category = preinstanced_parser_factory(
    'primary_category_id', DiscoveryCategory, DiscoveryCategory.general
)
put_primary_category_into = preinstanced_putter_factory('primary_category_id')
validate_primary_category = preinstanced_validator_factory('primary_category', DiscoveryCategory)

# sub_categories

parse_sub_categories = preinstanced_array_parser_factory('category_ids', DiscoveryCategory)
put_sub_categories_into = preinstanced_array_putter_factory('category_ids')
validate_sub_categories = preinstanced_array_validator_factory('sub_categories', DiscoveryCategory)



__all__ = ('ComponentMetadataStringSelect', )

from scarletio import copy_docs

from .fields import parse_options, put_options_into, validate_options
from .select_base import ComponentMetadataSelectBase


class ComponentMetadataStringSelect(ComponentMetadataSelectBase):
    """
    String select component metadata.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    options : `None`, `tuple` of ``StringSelectOption``
        Options of the select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    """
    __slots__ = ('options', )
    
    
    def __new__(
        cls,
        *,
        custom_id = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        options = ...,
        placeholder = ...,
    ):
        """
        Creates a new string select component metadata with the given parameters.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        options : `None`, `iterable` of ``StringSelectOption``, Optional (Keyword only)
            Options of the select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # options
        if options is ...:
            options = None
        else:
            options = validate_options(options)
        
        # Construct
        self = ComponentMetadataSelectBase.__new__(
            cls,
            custom_id = custom_id,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
        )
        self.options = options
        
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataSelectBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            custom_id = keyword_parameters.pop('custom_id', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            options = keyword_parameters.pop('options', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )
    
    
    @copy_docs(ComponentMetadataSelectBase._add_type_specific_repr_fields_into)
    def _add_type_specific_repr_fields_into(self, repr_parts):
        # options
        repr_parts.append(', options = ')
        options = self.options
        if (options is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
    
    
    @copy_docs(ComponentMetadataSelectBase.__hash__)
    def __hash__(self):
        hash_value = ComponentMetadataSelectBase.__hash__(self)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 12
            for option in options:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataSelectBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ComponentMetadataSelectBase._is_equal_same_type(self, other):
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataSelectBase.from_data)
    def from_data(cls, data):
        self = super(ComponentMetadataStringSelect, cls).from_data(data)
        self.options = parse_options(data)
        return self
    
    
    @copy_docs(ComponentMetadataSelectBase.to_data)
    def to_data(self, *, defaults = False):
        data =  ComponentMetadataSelectBase.to_data(self)
        
        put_options_into(self.options, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataSelectBase.copy)
    def copy(self):
        new = ComponentMetadataSelectBase.copy(self)
        
        # options
        options = self.options
        if (options is not None):
            options = tuple(option.copy() for option in options)
        new.options = options
        
        return new
    
    
    def copy_with(
        self,
        *,
        custom_id = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        options = ...,
        placeholder = ...,
    ):
        """
        Copies the channel select component metadata with the given fields.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        options : `None`, `iterable` of ``StringSelectOption``, Optional (Keyword only)
            Options of the select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # options
        if options is ...:
            options = self.options
            if (options is not None):
                options = tuple(option.copy() for option in options)
        else:
            options = validate_options(options)
        
        # Construct
        new = ComponentMetadataSelectBase.copy_with(
            self,
            custom_id = custom_id,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
        )
        new.options = options
        return new
    
    
    @copy_docs(ComponentMetadataSelectBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            custom_id = keyword_parameters.pop('custom_id', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            options = keyword_parameters.pop('options', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )



__all__ = ()

from os import environ as environmental_variables, environb as environmental_variables_binary
from os.path import dirname as get_directory_name, isfile as is_file, join as join_paths
from sys import _getframe as get_frame

from scarletio import RichAttributeErrorBaseType

from .parsing import parse_variables


MODULE_FRAME_NAME = '<module>'


def find_launched_location():
    """
    Finds the file that was launched.
    
    Returns
    -------
    location : `None`, `str`
    """
    frame = get_frame()
    
    last_named_location = None
    
    while True:
        frame = frame.f_back
        if frame is None:
            break
        
        # They might falsely name frames with auto generated functions, so check variables as well
        if (frame.f_code.co_name != MODULE_FRAME_NAME):
            continue
        
        frame_global = frame.f_globals
        # Try to not compare locals with globals if we are inside of a function.
        if (frame.f_locals is not frame_global):
            continue
        
        location = frame_global.get('__file__', None)
        if location is None:
            module_specification = frame_global.get('__spec__', None)
            if module_specification is not None:
                location = module_specification.origin
        
        if location is not None:
            last_named_location = location
    
    return last_named_location


def find_dot_env_file():
    """
    Tries to find dotenv location to load.
    
    returns
    -------
    location : `None`, `str`
    """
    launched_location = find_launched_location()
    if (launched_location is not None):
        file_path = join_paths(get_directory_name(launched_location), '.env')
        if is_file(file_path):
            return file_path


class DotEnvResult(RichAttributeErrorBaseType):
    """
    Represents result of loading a `.env` file or just content.
    
    Attributes
    ----------
    file_path : `str`
        Path to the loaded file.
    parser_failure_info : `None`, ``ParserFailureInfo``
        Failure info if parsing failed.
    variables : `dict<str, None | str>`
        The loaded environmental variables.
    """
    __slots__ = ('file_path', 'parser_failure_info', 'variables')
    
    def __new__(cls, variables, parser_failure_info, file_path):
        """
        Creates a new dot-env result.
        
        Parameters
        ----------
        variables : `dict<str, None | str>`
            The loaded environmental variables.
        parser_failure_info : `None`, ``ParserFailureInfo``
            Failure info if parsing failed.
        file_path : `str`
            Path to the loaded file.
        """
        self = object.__new__(cls)
        self.variables = variables
        self.parser_failure_info = parser_failure_info
        self.file_path = file_path
        return self
    
    
    def insert_to_environmental_variables(self):
        """
        Inserts the parsed variables into the environmental ones. If the variable already exists, will not overwrite it.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        for key, value in self.variables.items():
            if value is None:
                value = ''
            
            environmental_variables.setdefault(key, value)
            environmental_variables_binary.setdefault(key.encode(), value.encode())
        
        return self
    
    
    def raise_if_failed(self):
        """
        Raises an exception with a reason why loading failed.
        
        Returns
        -------
        self : `instance<type<self>>`
        
        Raises
        ------
        SyntaxError
        """
        parser_failure_info = self.parser_failure_info
        if (parser_failure_info is not None):
            raise SyntaxError(
                parser_failure_info.get_error_message(),
                (
                    self.file_path,
                    parser_failure_info.line_index + 1,
                    parser_failure_info.index + 1,
                    parser_failure_info.line,
                ),
            )
        
        return self
    
    
    def __repr__(self):
        """Returns the dot env result's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' variables: ')
        repr_parts.append(repr(len(self.variables)))
        
        parser_failure_info = self.parser_failure_info
        if (parser_failure_info is not None):
            repr_parts.append(', parser_failure_info = ')
            repr_parts.append(repr(parser_failure_info))
        
        file_path = self.file_path
        if (file_path is not None):
            repr_parts.append(' file_path = ')
            repr_parts.append(repr(file_path))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


def load_dot_env(value, file_path = None):
    """
    Loads the `.env` variables from the given content
    
    Parameters
    ----------
    value : `str`
        The value to parse.
    file_path : `None`, `str`
        File path to show u
    
    Returns
    -------
    dot_env_result : ``DotEnvResult``
    """
    variables, parser_failure_info = parse_variables(value)
    return DotEnvResult(variables, parser_failure_info, file_path)


def load_dot_env_from_file(file_path):
    """
    Loads the `.env` file from the given path.
    
    Parameters
    ----------
    file_path : `str`
        The path of the file to load.
    
    Returns
    -------
    dot_env_result : ``DotEnvResult``
    """
    with open(file_path, 'r') as file:
        value = file.read()
    
    return load_dot_env(value, file_path)



__all__ = ('Attachment', )

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_content_type, parse_description, parse_duration, parse_height, parse_id, parse_name, parse_proxy_url,
    parse_size, parse_temporary, parse_url, parse_waveform, parse_width, put_content_type_into, put_description_into,
    put_duration_into, put_height_into, put_id_into, put_name_into, put_proxy_url_into, put_size_into,
    put_temporary_into, put_url_into, put_waveform_into, put_width_into, validate_content_type, validate_description,
    validate_duration, validate_height, validate_id, validate_name, validate_proxy_url, validate_size,
    validate_temporary, validate_url, validate_waveform, validate_width
)


PRECREATE_FIELDS = {
    'content_type': ('content_type', validate_content_type),
    'description': ('description', validate_description),
    'duration': ('duration', validate_duration),
    'height': ('height', validate_height),
    'name': ('name', validate_name),
    'proxy_url': ('proxy_url', validate_proxy_url),
    'size': ('size', validate_size),
    'temporary': ('temporary', validate_temporary),
    'url': ('url', validate_url),
    'waveform': ('waveform', validate_waveform),
    'width': ('width', validate_width),
}


class Attachment(DiscordEntity):
    """
    Represents an attachment of a ``Message``.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the attachment.
    
    content_type : `None`, `str`
        The attachment's media type.
    
    description : `None`, `str`
        Description for the file.
        
        > Max 1024 characters.
    
    duration : `float`
        The attachment's duration in seconds. Applicable for voice messages only.
        
        > Defaults to `0.0`.
    
    height : `int`
        The height of the attachment if applicable.
        
        > Defaults to `0`.
    
    name : `str`
        The name of the attachment.
    
    proxy_url : `str`
        Proxied url of the attachment.
    
    size : `int`
        The attachment's size in bytes.
    
    temporary : `bool`
        Whether the attachment is temporary and is removed after a set period of time.
        
        Temporary attachments are guaranteed to be available as long as their message itself exists.
        
        > Defaults to `False`.
    
    url : `str`
        The attachment's url.
    
    waveform : `None`, `str`
        Base64 encoded bytearray representing a sampled waveform. Applicable for voice messages only.
        
        > Defaults to `None`.
    
    width : `int`
        The attachment's width if applicable.
        
        > Defaults to `0`.
    """
    __slots__ = (
        'content_type', 'description', 'duration', 'height', 'name', 'proxy_url', 'size', 'temporary', 'url',
        'waveform', 'width'
    )
    
    def __new__(
        cls,
        *,
        content_type = ...,
        description = ...,
        duration = ...,
        height = ...,
        name = ...,
        size = ...,
        temporary = ...,
        url = ...,
        waveform = ...,
        width = ...,
    ):
        """
        Creates a new partial attachment.
        
        Parameters
        ----------
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        duration : `float`, Optional (Keyword only)
            The length of the file in seconds.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
        
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # content_type
        if content_type is ...:
            content_type = None
        else:
            content_type = validate_content_type(content_type)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # duration
        if duration is ...:
            duration = 0.0
        else:
            duration = validate_duration(duration)
        
        # height
        if height is ...:
            height = 0
        else:
            height = validate_height(height)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # size
        if size is ...:
            size = 0
        else:
            size = validate_size(size)
        
        # temporary
        if temporary is ...:
            temporary = False
        else:
            temporary = validate_temporary(temporary)
        
        # url
        if url is ...:
            url = ''
        else:
            url = validate_url(url)
        
        # waveform
        if waveform is ...:
            waveform = None
        else:
            waveform = validate_waveform(waveform)
        
        # width
        if width is ...:
            width = 0
        else:
            width = validate_width(width)
        
        # Construct
        
        self = object.__new__(cls)
        self.content_type = content_type
        self.description = description
        self.duration = duration
        self.height = height
        self.id = 0
        self.name = name
        self.proxy_url = None
        self.size = size
        self.temporary = temporary
        self.url = url
        self.waveform = waveform
        self.width = width
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an attachment object from the attachment data included inside of a ``Message`'s.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received attachment data.
        """
        self = object.__new__(cls)
        self.content_type = parse_content_type(data)
        self.description = parse_description(data)
        self.duration = parse_duration(data)
        self.height = parse_height(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.proxy_url = parse_proxy_url(data)
        self.size = parse_size(data)
        self.temporary = parse_temporary(data)
        self.url = parse_url(data)
        self.waveform = parse_waveform(data)
        self.width = parse_width(data)
        return self
    
    
    def __repr__(self):
        """Returns the representation of the attachment."""
        repr_parts = [
            '<', self.__class__.__name__,
        ]
        
        attachment_id = self.id
        if attachment_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(attachment_id))
            
            if self.temporary:
                repr_parts.append(' (temporary)')
            
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # Extra if audio
        duration = self.duration
        if duration:
            repr_parts.append(', duration = ')
            repr_parts.append(format(duration, '.02f'))
        
        # Extra if image
        width = self.width
        height = self.height
        if width and height:
            repr_parts.append(', size = ')
            repr_parts.append(repr(width))
            repr_parts.append('x')
            repr_parts.append(repr(height))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two attachments are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # shortcut
        if self is other:
            return True
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            # Only compare `.id` and `.proxy_url` if both attachment is not partial.
            
            # id
            if self.id != other.id:
                return False
            
            # proxy_url
            if self.proxy_url != other.proxy_url:
                return False
            
        # content_type
        if self.content_type != other.content_type:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # duration
        if self.duration != other.duration:
            return False
        
        # height
        if self.height != other.height:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # size
        if self.size != other.size:
            return False
        
        # temporary
        if self.temporary != other.temporary:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        # waveform
        if self.waveform != other.waveform:
            return False
        
        # width
        if self.width != other.width:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the attachment."""
        hash_value = 0
        
        # content_type
        content_type = self.content_type
        if (content_type is not None):
            hash_value ^= hash(content_type)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # duration
        duration = self.duration
        if duration:
            hash_value ^= hash(duration)
        
        # height
        hash_value ^= self.height
        
        # id
        attachment_id = self.id
        if attachment_id:
            hash_value ^= attachment_id
            
            # proxy_url
            # Add proxy url to hash only if we have attachment id.
            proxy_url = self.proxy_url
            if (proxy_url is not None):
                hash_value ^= hash(proxy_url)
        
        # name
        name = self.name
        if (description is None) or (name != description):
            hash_value ^= hash(name)
        
        # size
        hash_value ^= self.size << 17
        
        # temporary
        hash_value ^= self.temporary << 16
        
        # url
        hash_value ^= hash(self.url)
        
        # waveform
        waveform = self.waveform
        if (waveform is not None):
            hash_value ^= hash(waveform)
        
        # width
        hash_value ^= self.width << 8
        
        return hash_value
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Tries to convert the attachment back to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`)
        """
        data = {}
        
        # content_type
        put_content_type_into(self.content_type, data, defaults)
        
        # description
        put_description_into(self.description, data, defaults)
        
        # duration
        put_duration_into(self.duration, data, defaults)
        
        # height
        put_height_into(self.height, data, defaults)
        
        # id
        if include_internals:
            put_id_into(self.id, data, defaults)
        
        # name
        put_name_into(self.name, data, defaults)
        
        # proxy_url
        if include_internals:
            put_proxy_url_into(self.proxy_url, data, defaults)
        
        # size
        put_size_into(self.size, data, defaults)
        
        # temporary
        put_temporary_into(self.temporary, data, defaults)
        
        # url
        put_url_into(self.url, data, defaults)
        
        # waveform
        put_waveform_into(self.waveform, data, defaults)
        
        # width
        put_width_into(self.width, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Returns a copy of the attachment.
        
        > The copy will not include the internal fields of it.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.content_type = self.content_type
        new.description = self.description
        new.duration = self.duration
        new.height = self.height
        new.id = 0
        new.name = self.name
        new.proxy_url = None
        new.size = self.size
        new.temporary = self.temporary
        new.url = self.url
        new.waveform = self.waveform
        new.width = self.width
        return new
    
    
    def copy_with(
        self,
        *,
        content_type = ...,
        description = ...,
        duration = ...,
        height = ...,
        name = ...,
        size = ...,
        temporary = ...,
        url = ...,
        waveform = ...,
        width = ...,
    ):
        """
        Returns a copy of the attachment. The attributes of the copy are modified based on the given values.
        
        > The copy will not include the internal fields of it.
        
        Parameters
        ----------
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            Description for the file.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
        
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # content_type
        if content_type is ...:
            content_type = self.content_type
        else:
            content_type = validate_content_type(content_type)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # duration
        if duration is ...:
            duration = self.duration
        else:
            duration = validate_duration(duration)
        
        # height
        if height is ...:
            height = self.height
        else:
            height = validate_height(height)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # size
        if size is ...:
            size = self.size
        else:
            size = validate_size(size)
        
        # temporary
        if temporary is ...:
            temporary = self.temporary
        else:
            temporary = validate_temporary(temporary)
        
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        # waveform
        if waveform is ...:
            waveform = self.waveform
        else:
            waveform = validate_waveform(waveform)
        
        # width
        if width is ...:
            width = self.width
        else:
            width = validate_width(width)
        
        # Construct
        
        new = object.__new__(type(self))
        new.content_type = content_type
        new.description = description
        new.duration = duration
        new.height = height
        new.id = 0
        new.name = name
        new.proxy_url = None
        new.size = size
        new.temporary = temporary
        new.url = url
        new.waveform = waveform
        new.width = width
        return new
    
    
    @classmethod
    def precreate(
        cls,
        attachment_id,
        **keyword_parameters,
    ):
        """
        Precreates an attachment. Since attachments are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        attachment_id : `int`
            The attachment's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the attachment's fields should be set.
        
        Other Parameters
        ----------------
        content_type : `None`, `str`, Optional (Keyword only)
            The attachment's media type.
        
        description : `None`, `str`, Optional (Keyword only)
            Description for the file.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        height : `int`, Optional (Keyword only)
            The height of the attachment if applicable.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        size : `int`, Optional (Keyword only)
            The attachment's size in bytes.
        
        proxy_url : `None`, `str`, Optional (Keyword only)
            Proxied url of the attachment.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the attachment is temporary and is removed after a set period of time.
            
        url : `str`, Optional (Keyword only)
            The attachment's url.
        
        waveform : `None`, `str`, Optional (Keyword only)
            Base64 encoded bytearray representing a sampled waveform.
        
        width : `int`, Optional (Keyword only)
            The attachment's width if applicable.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        attachment_id = validate_id(attachment_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.content_type = None
        self.description = None
        self.duration = 0.0
        self.height = 0
        self.id = attachment_id
        self.name = ''
        self.proxy_url = None
        self.size = 0
        self.temporary = False
        self.url = ''
        self.waveform = None
        self.width = 0
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self



__all__ = ()

from ...activity import Activity

from ..activity_update import ActivityUpdate


def _validate_entity_list(entity_list_name, entity_type, entity_list):
    """
    Validates an entity list. Used by `validate_added` and `validate_removed`.
    
    Parameters
    ----------
    entity_list_name : `str`
        The name of the entity list.
    entity_list : `None`, `iterable` of `entity_type`
        The entity list to validate.
    
    Returns
    -------
    entity_list : `None`, `list` of `entity_type`
    
    Raises
    ------
    TypeError
        - If `entity_list`'s type is incorrect.
    """
    if entity_list is None:
        return None
    
    if getattr(entity_list, '__iter__', None) is None:
        raise TypeError(
            f'`{entity_list_name}` can be `None`, `iterable` of `{entity_type.__name__}`, got '
            f'{entity_list.__class__.__name__}; {entity_list!r}.'
        )
    
    entities_validated = None
    
    for entity in entity_list:
        if not isinstance(entity, entity_type):
            raise TypeError(
                f'`{entity_list_name}` elements can be `{entity_type.__name__}`, got '
                f'{entity.__class__.__name__}; {entity!r}; {entity_list_name} = {entity_list!r}'
            )
        
        if entities_validated is None:
            entities_validated = []
        
        entities_validated.append(entity)
    
    return entities_validated


def validate_added(added):
    """
    Validates the `added` field of an activity change.
    
    Parameters
    ----------
    added : `None`, `iterable` of ``Activity``
        Added activity to validate.
    
    Returns
    -------
    added : `None`, `list` of ``Activity``
    
    Raises
    ------
    TypeError
        - If `added`'s type is incorrect.
    """
    return _validate_entity_list('added', Activity, added)


def validate_updated(updated):
    """
    Validates the `updated` field of an activity change.
    
    Parameters
    ----------
    updated : `None`, `iterable` of ``ActivityUpdate``
        Added activity to validate.
    
    Returns
    -------
    updated : `None`, `list` of ``ActivityUpdate``
    
    Raises
    ------
    TypeError
        - If `updated`'s type is incorrect.
    """
    return _validate_entity_list('updated', ActivityUpdate, updated)


def validate_removed(removed):
    """
    Validates the `removed` field of an activity change.
    
    Parameters
    ----------
    removed : `None`, `iterable` of ``Activity``
        Added activity to validate.
    
    Returns
    -------
    removed : `None`, `list` of ``Activity``
    
    Raises
    ------
    TypeError
        - If `removed`'s type is incorrect.
    """
    return _validate_entity_list('removed', Activity, removed)



__all__ = ('SoundboardSoundsEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS

from .fields import parse_guild_id, parse_sounds, put_guild_id_into, put_sounds_into, validate_guild_id, validate_sounds


class SoundboardSoundsEvent(EventBase):
    """
    Represents a soundboard sound event sent to a voice channel.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier that the event represents.
    sounds : `None`, `tuple` of ``SoundboardSound``
        The responded sounds.
    """
    __slots__ = ('guild_id', 'sounds')
    
    def __new__(
        cls,
        *,
        guild_id = ...,
        sounds = ...,
    ):
        """
        Creates a self soundboard sounds event
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier that the event represents.
        sounds : `None`, `iterable` of ``SoundboardSound``, Optional (Keyword only)
            The responded sounds.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sounds
        if sounds is ...:
            sounds = None
        else:
            sounds = validate_sounds(sounds)
        
        # Construct
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.sounds = sounds
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a soundboard sounds event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Voice channel effect event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        guild_id = parse_guild_id(data)
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            old_sounds = [*guild.iter_soundboard_sounds()]
            guild.soundboard_sounds = None
            guild.soundboard_sounds_cached = True
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.sounds = parse_sounds(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the soundboard sounds event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_guild_id_into(self.guild_id, data, defaults)
        put_sounds_into(self.sounds, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', sounds = [')
        
        sounds = self.sounds
        if (sounds is not None):
            index = 0
            length = len(sounds)
            
            while True:
                sound = sounds[index]
                repr_parts.append(repr(sound))
                
                index += 1
                if index == length:
                    break
                    
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.sounds
        
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # sounds
        if self.sounds != other.sounds:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # guild_id
        hash_value ^= self.guild_id
        
        # sounds
        sounds = self.sounds
        if (sounds is not None):
            hash_value ^= len(sounds)
            
            for sound in sounds:
                hash_value ^= hash(sound)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the soundboard sounds event
        
        Returns
        -------
        new : `instance<type<new>>`
        """
        new = object.__new__(type(self))
        new.guild_id = self.guild_id
        sounds = self.sounds
        if (sounds is not None):
            sounds = (*(sound for sound in sounds),)
        new.sounds = sounds
        return new
    
    
    def copy_with(
        self,
        *,
        guild_id = ...,
        sounds = ...,
    ):
        """
        Copies new soundboard sounds event with the given fields.
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier that the event represents.
        sounds : `None`, `iterable` of ``SoundboardSound``, Optional (Keyword only)
            The responded sounds.
        
        Returns
        -------
        new : `instance<type<new>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sounds
        if sounds is ...:
            sounds = self.sounds
            if (sounds is not None):
                sounds = (*(sound for sound in sounds),)
        else:
            sounds = validate_sounds(sounds)
        
        # Construct
        new = object.__new__(type(self))
        new.guild_id = guild_id
        new.sounds = sounds
        return new
    
    
    @property
    def guild(self):
        """
        Returns the soundboard sound event's guild. If the guild is not cached returns `None`.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_sounds(self):
        """
        Iterates over the sounds of the soundboard sounds event.
        
        This method is an iterable generator
        
        Yields
        ------
        sound : ``SoundboardSound``
        """
        sounds = self.sounds
        if (sounds is not None):
            yield from sounds



__all__ = ()

from ..audit_log_change import AuditLogChange
from ..audit_log_role import AuditLogRole

from .shared import convert_icon, convert_nothing, convert_timestamp


def convert_roles(name, data):
    roles = tuple(AuditLogRole(role_data) for role_data in data['new_value'])
    
    if name == '$add':
        before = None
        after = roles
    else:
        before = roles
        after = None
    
    return AuditLogChange('roles', before, after)


def convert_timestamp__timed_out_until(name, data):
    return convert_timestamp('timed_out_until', data)


USER_CONVERTERS = {
    '$add': convert_roles,
    '$remove': convert_roles,
    'avatar_hash': convert_icon,
    'communication_disabled_until': convert_timestamp__timed_out_until,
    'deaf': convert_nothing,
    'mute': convert_nothing,
    'nick': convert_nothing,
    'pending': convert_nothing,
}



__all__ = ()

import warnings

from scarletio import Compound


from ...allowed_mentions import parse_allowed_mentions
from ...application import Application
from ...bases import maybe_snowflake
from ...component import InteractionForm
from ...http import DiscordHTTPClient
from ...interaction import InteractionEvent, InteractionResponseContext, InteractionResponseType, InteractionType

from ...message import Message, MessageFlag
from ...message.message.utils import try_resolve_interaction_message

from ..functionality_helpers import application_command_autocomplete_choice_parser
from ..request_helpers import add_file_to_message_data, get_components_data, validate_content_and_embed


MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY = MessageFlag().update_by_keys(invoking_user_only = True)
MESSAGE_FLAG_VALUE_SILENT = MessageFlag().update_by_keys(silent = True)
MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS = MessageFlag().update_by_keys(embeds_suppressed = True)


def _assert__interaction_event_type(interaction_event):
    """
    Asserts whether the interaction event's type is correct.
    
    Parameters
    ----------
    interaction_event : ``InteractionEvent``
        Respective interaction event.
    
    Raises
    ------
    AssertionError
        - If `interaction_event` is not ``InteractionEvent``.
    """
    if not isinstance(interaction_event, InteractionEvent):
        raise AssertionError(
            f'`interaction` can be `{InteractionEvent.__name__}`, got '
            f'{interaction_event.__class__.__name__}; {interaction_event!r}.'
        )
    
    return True


def _assert__application_id(application_id):
    """
    Asserts whether the client's application is synced by checking the `application_id`'s value.
    
    Parameters
    ----------
    application_id : `int`
        Client's application's id.
    
    Raises
    ------
    AssertionError
        - If the client's application is not yet synced..
    """
    if application_id == 0:
        raise AssertionError(
            'The client\'s application is not yet synced.'
        )
    
    return True


def _assert__form(form):
    """
    Asserts whether the form's type is correct.
    
    Parameters
    ----------
    form : ``InteractionForm``
        Respective form.
    
    Raises
    ------
    AssertionError
        - If `form` is not ``InteractionForm``.
    """
    if not isinstance(form, InteractionForm):
        raise AssertionError(
            f'`form` can be `{InteractionForm.__name__}`, got '
            f'{form.__class__.__name__}; {form!r}.'
        )
    
    return True


def _assert__show_for_invoking_user_only(show_for_invoking_user_only):
    """
    Asserts whether the given `show_for_invoking_user_only`'s type is correct.
    
    Parameters
    ----------
    show_for_invoking_user_only : ``bool``
        Whether the sent message should only be shown to the invoking user.
    
    Raises
    ------
    AssertionError
        - If `show_for_invoking_user_only` is not ``bool``.
    """
    if not isinstance(show_for_invoking_user_only, bool):
        raise AssertionError(
            f'`show_for_invoking_user_only` can be `bool`, got '
            f'{show_for_invoking_user_only.__class__.__name__}; {show_for_invoking_user_only!r}.'
        )
    
    return True


def _assert__silent(silent):
    """
    Asserts whether the given `silent`'s type is correct.
    
    Parameters
    ----------
    silent : ``bool``
        Whether the message should be delivered silently.
    
    Raises
    ------
    AssertionError
        - If `silent` is not ``bool``.
    """
    if not isinstance(silent, bool):
        raise AssertionError(
            f'`silent` can be `bool`, got '
            f'{silent.__class__.__name__}; {silent!r}.'
        )
    
    return True


def _assert__suppress_embeds(suppress_embeds):
    """
    Asserts whether the given `suppress_embeds`'s type is correct.
    
    Parameters
    ----------
    suppress_embeds : ``bool``
        Whether the message's embeds should be suppressed initially.
    
    Raises
    ------
    AssertionError
        - If `suppress_embeds` is not ``bool``.
    """
    if not isinstance(suppress_embeds, bool):
        raise AssertionError(
            f'`suppress_embeds` can be `bool`, got '
            f'{suppress_embeds.__class__.__name__}; {suppress_embeds!r}.'
        )
    
    return True


def _assert__tts(tts):
    """
    Asserts whether the given `tts`'s type is correct.
    
    Parameters
    ----------
    tts : ``bool``
        Whether the message is text-to-speech.
    
    Raises
    ------
    AssertionError
        - If `tts` is not ``bool``.
    """
    if not isinstance(tts, bool):
        raise AssertionError(
            f'`tts` can be `bool`, got '
            f'{tts.__class__.__name__}; {tts!r}.'
        )
    
    return True


class ClientCompoundInteractionEndpoints(Compound):
    
    application : Application
    http : DiscordHTTPClient
    
    
    async def interaction_application_command_acknowledge(
        self, interaction_event, wait = True, *, show_for_invoking_user_only = False
    ):
        """
        Acknowledges the given application command interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge
        wait : `bool` = `True`, Optional
            Whether the interaction should be ensured asynchronously.
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to `False`.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code=10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not ack twice
        if not interaction_event.is_unanswered():
            return
        
        data = {'type': InteractionResponseType.source.value}
        
        if show_for_invoking_user_only:
            data['data'] = {'flags': MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY}
        
        context = InteractionResponseContext(interaction_event, True, show_for_invoking_user_only)
        coroutine = self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
        
        if wait:
            async with context:
                await coroutine
        else:
            await context.ensure(coroutine)
    
    
    async def interaction_application_command_autocomplete(self, interaction_event, choices):
        """
        Forwards auto completion choices for the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge
        choices : `None`, `iterable` of `str`
            Choices to show for the user.
        
        Raises
        ------
        TypeError
            If `choice` is neither `None` nor `iterable`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code=10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not auto complete twice
        if not interaction_event.is_unanswered():
            return
        
        choices = application_command_autocomplete_choice_parser(choices)
        
        data = {
            'type': InteractionResponseType.application_command_autocomplete_result.value,
            'data': {
                'choices': choices,
            },
        }
        
        async with InteractionResponseContext(interaction_event, True, False):
            await self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
    
    
    async def interaction_form_send(self, interaction_event, form):
        """
        Responds on an interaction with a form.
        
        This function is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to respond to.
        form : ``InteractionForm``
            The to respond with.
        
        Raises
        ------
        RuntimeError
            If cannot respond with a form on the given `interaction`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Discord do not returns message data, so the method cannot return a ``Message`` either.
        
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code=10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        if not interaction_event.is_unanswered():
            warnings.warn(
                (
                    f'`{self.__class__.__name__}.interaction_response_form` called on an interaction already '
                    f'acknowledged / answered: {interaction_event!r}. Returning `None`.'
                ),
                ResourceWarning,
                stacklevel = 2,
            )
            
            return None
        
        assert _assert__form(form)
        
        if (
            (interaction_event.type is not InteractionType.application_command) and
            (interaction_event.type is not InteractionType.message_component)
        ):
            raise RuntimeError(
                f'Only `application_command` and `message_component` interactions can be answered with '
                f'form, got `{interaction_event.type.name}`; {interaction_event!r}; form={form!r}.'
            )
        
        # Build payload
        data = {
            'data': form.to_data(),
            'type': InteractionResponseType.form.value,
        }
        
        async with InteractionResponseContext(interaction_event, False, True):
            await self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
        
        return None
    
    
    async def interaction_response_message_create(
        self,
        interaction_event,
        content = None,
        *,
        allowed_mentions = ...,
        components = None,
        embed = None,
        show_for_invoking_user_only = False,
        silent = False,
        suppress_embeds = False,
        tts = False,
    ):
        """
        Sends an interaction response. After receiving an ``InteractionEvent``, you should acknowledge it within
        `3` seconds to perform followup actions.
        
        Not like ``.message_create``, this endpoint can be called without any content to still acknowledge the
        interaction event. This method also wont return a ``Message`` object (thank to Discord), but at least
        ``.interaction_followup_message_create`` will. To edit or delete this message, you can use
        ``.interaction_response_message_edit`` and ``interaction_response_message_delete``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to respond to.
        
        content : `None`, `str`, ``Embed``, `object` = `None`, Optional
            The interaction response's content if given. If given as `str` or empty string, then no content will be
            sent, meanwhile if any other non `str`, ``Embed`` is given, then will be casted to string.
            
            If given as ``Embed``, then is sent as the message's embed.
        
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of \
                ``Component``) = `None`, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        
        embed : `None`, ``Embed``, `list` of ``Embed`` = `None`, Optional (Keyword only)
            The embedded content of the interaction response.
            
            If `embed` and `content` parameters are both given as ``Embed``, then `AssertionError` is
            raised.
        
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to `False`.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        Raises
        ------
        TypeError
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``Embed`` nor as `list`, `tuple` of ``Embed``-s.
            - If `content` parameter was given as ``Embed``, meanwhile `embed` parameter was given as well.
            - If `components` type is incorrect.
        ValueError
            If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Discord do not returns message data, so the method cannot return a ``Message`` either.
        
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code=10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        content, embed = validate_content_and_embed(content, embed, False)
        components = get_components_data(components, False)
        
        assert _assert__show_for_invoking_user_only(show_for_invoking_user_only)
        assert _assert__silent(silent)
        assert _assert__suppress_embeds(suppress_embeds)
        assert _assert__tts(tts)
        
        # Build payload
        
        message_data = {}
        contains_content = False
        
        if (content is not None):
            message_data['content'] = content
            contains_content = True
        
        if (embed is not None):
            message_data['embeds'] = [embed.to_data() for embed in embed]
            contains_content = True
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not None):
            message_data['components'] = components
            contains_content = True
        
        if tts:
            message_data['tts'] = True
        
        if message_data:
            is_deferring = False
        else:
            is_deferring = True
        
        flags = (
            (MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY if show_for_invoking_user_only else 0) |
            (MESSAGE_FLAG_VALUE_SILENT if silent else 0) |
            (MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS if suppress_embeds else 0)
        )
        if flags:
            message_data['flags'] = flags
            contains_content = True
        
        data = {}
        if contains_content:
            data['data'] = message_data
        
        if is_deferring:
            response_type = InteractionResponseType.source
        else:
            response_type = InteractionResponseType.message_and_source
        
        data['type'] = response_type.value
        
        async with InteractionResponseContext(interaction_event, is_deferring, show_for_invoking_user_only):
            await self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
        
        # No message data is returned by Discord, return `None`.
        return None
    
    
    async def interaction_component_acknowledge(self, interaction_event, wait = True):
        """
        Acknowledges the given component interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge
        wait : `bool` = `True`, Optional
            Whether the interaction should be ensured asynchronously.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code=10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not ack twice
        if not interaction_event.is_unanswered():
            return
        
        data = {'type': InteractionResponseType.component.value}
        
        context = InteractionResponseContext(interaction_event, True, False)
        coroutine = self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
        
        
        if wait:
            async with context:
                await coroutine
        else:
            await context.ensure(coroutine)
    
    
    async def interaction_response_message_edit(
        self,
        interaction_event,
        content = ...,
        *,
        allowed_mentions = ...,
        components = ...,
        embed = ...,
        file = ...,
    ):
        """
        Edits the given `interaction`'s source response. If the source interaction event was only deferred, this call
        will send the message as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be edited.
        
        content : `None`, `str`, ``Embed``, `object`, Optional
            The new content of the message.
            
            If given as ``Embed``, then the message's embeds will be edited with it.
        
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of \
                ``Component``), Optional (Keyword only)
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        embed : `None`, ``Embed``, `list` of ``Embed``, Optional (Keyword only)
            The new embedded content of the message. By passing it as `None`, you can remove the old.
            
            If `embed` and `content` parameters are both given as  ``Embed``, then `AssertionError` is
            raised.
        
        file : `None`, `object`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        Raises
        ------
        TypeError
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``Embed`` nor as `list`, `tuple` of ``Embed``-s.
            - If `content` parameter was given as ``Embed``, meanwhile `embed` parameter was given as well.
        ValueError
            If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Cannot editing interaction messages, which were created with `show_for_invoking_user_only=True`:
        
        ```
        DiscordException Not Found (404), code=10008: Unknown Message
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        content, embed = validate_content_and_embed(content, embed, True)
        
        components = get_components_data(components, True)
        
        # Build payload
        message_data = {}
        
        # Discord docs say, content can be nullable, but nullable content is just ignored.
        if (content is not ...):
            message_data['content'] = content
        
        if (embed is not ...):
            if (embed is not None):
                embed = [embed.to_data() for embed in embed]
            
            message_data['embeds'] = embed
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not ...):
            message_data['components'] = components
        
        message_data = add_file_to_message_data(message_data, file, True, True)
        
        async with InteractionResponseContext(interaction_event, False, False):
            await self.http.interaction_response_message_edit(
                application_id, interaction_event.id, interaction_event.token, message_data
            )
    
    
    async def interaction_component_message_edit(
        self,
        interaction_event,
        content = ...,
        *,
        embed = ...,
        allowed_mentions = ...,
        components = ...,
    ):
        """
        Edits the given component interaction's source message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be edited.
        
        content : `None`, `str`, ``Embed``, `object`, Optional
            The new content of the message.
            
            If given as ``Embed``, then the message's embeds will be edited with it.
        
        embed : `None`, ``Embed``, `list` of ``Embed``, Optional (Keyword only)
            The new embedded content of the message. By passing it as `None`, you can remove the old.
            
            If `embed` and `content` parameters are both given as  ``Embed``, then `AssertionError` is
            raised.
        
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        Raises
        ------
        TypeError
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``Embed`` nor as `list`, `tuple` of ``Embed``-s.
            - If `content` parameter was given as ``Embed``, meanwhile `embed` parameter was given as well.
        ValueError
            If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        content, embed = validate_content_and_embed(content, embed, True)
        
        components = get_components_data(components, True)
        
        # Build payload
        message_data = {}
        
        if (content is not ...):
            message_data['content'] = content
        
        if (embed is not ...):
            if (embed is not None):
                embed = [embed.to_data() for embed in embed]
            
            message_data['embeds'] = embed
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not ...):
            message_data['components'] = components
        
        data = {
            'data': message_data,
            'type': InteractionResponseType.component_message_edit.value,
        }
        
        
        async with InteractionResponseContext(interaction_event, False, False):
            await self.http.interaction_response_message_create(interaction_event.id, interaction_event.token, data)
    
    
    async def interaction_response_message_delete(self, interaction_event):
        """
        Deletes the given `interaction`'s source response message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be deleted.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        await self.http.interaction_response_message_delete(application_id, interaction_event.id, interaction_event.token)
    
    
    async def interaction_response_message_get(self, interaction_event):
        """
        Gets the given `interaction`'s source response message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be deleted.
        
        Returns
        -------
        message : ``Message``
            The created message.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        
        message_data = await self.http.interaction_response_message_get(
            application_id, interaction_event.id, interaction_event.token
        )
        
        return Message.from_data(message_data)
    
    
    async def interaction_followup_message_create(
        self,
        interaction_event,
        content = None,
        *,
        allowed_mentions = ...,
        components = None,
        embed = None,
        file = None,
        show_for_invoking_user_only = False,
        silent = False,
        suppress_embeds = False,
        tts = False,
    ):
        """
        Sends a followup message with the given interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to create followup message with.
        
        content : `None`, `str`, ``Embed``, `object` = `True`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str`, ``Embed`` is given, then will be casted to string.
            
            If given as ``Embed``, then is sent as the message's embed.
        
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        

        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of \
                ``Component``) = `None`, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        
        embed : `None`, ``Embed``, `list` of ``Embed`` = `None`, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``Embed``, then `TypeError` is raised.
        
        file : `None`, `object` = `None`, Optional
            A file to send. Check ``create_file_form`` for details.
        
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to `False`.
            
            Invoking user only messages can have attachments too. These attachments are purged by Discord after a set
            amount of time (2 weeks), so do not rely on reusing their url.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech. Defaults to `False`.
        
        
        Returns
        -------
        message : `None`, ``Message``
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``Embed`` nor as `list`, `tuple` of ``Embed``-s.
            - `content` parameter was given as ``Embed``, meanwhile `embed` parameter was given as well.
            - If invalid file type would be sent.
            - If `components` type is incorrect.
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        content, embed = validate_content_and_embed(content, embed, False)
        components = get_components_data(components, False)
        
        assert _assert__show_for_invoking_user_only(show_for_invoking_user_only)
        assert _assert__silent(silent)
        assert _assert__suppress_embeds(suppress_embeds)
        assert _assert__tts(tts)
        
        # Build payload
        
        message_data = {}
        contains_content = False
        
        if (content is not None):
            message_data['content'] = content
            contains_content = True
        
        if (embed is not None):
            message_data['embeds'] = [embed.to_data() for embed in embed]
            contains_content = True
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not None):
            message_data['components'] = components
            contains_content = True
        
        if tts:
            message_data['tts'] = True
        
        flags = (
            (MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY if show_for_invoking_user_only else 0) |
            (MESSAGE_FLAG_VALUE_SILENT if silent else 0) |
            (MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS if suppress_embeds else 0)
        )
        if flags:
            message_data['flags'] = flags
        
        message_data = add_file_to_message_data(message_data, file, contains_content, False)
        if message_data is None:
            return
        
        async with InteractionResponseContext(interaction_event, False, show_for_invoking_user_only):
            message_data = await self.http.interaction_followup_message_create(
                application_id, interaction_event.id, interaction_event.token, message_data
            )
        
        message = interaction_event.channel._create_new_message(message_data)
        try_resolve_interaction_message(message, interaction_event)
        return message
    
    
    async def interaction_followup_message_edit(
        self,
        interaction_event,
        message,
        content = ...,
        *,
        embed = ...,
        file = ...,
        allowed_mentions = ...,
        components = ...,
    ):
        """
        Edits the given interaction followup message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        
        message : ``Message``, `int`
            The interaction followup's message to edit.
        
        content : `None`, `str`, ``Embed``, `object`, Optional
            The new content of the message.
            
            If given as `str` then the message's content will be edited with it. If given as any non ``Embed``
            instance, then it will be cased to string first.
            
            By passing it as empty string, you can remove the message's content.
            
            If given as ``Embed``, then the message's embeds will be edited with it.
        
        embed : `None`, ``Embed``, `list` of ``Embed``, Optional (Keyword only)
            The new embedded content of the message. By passing it as `None`, you can remove the old.
            
            If `embed` and `content` parameters are both given as  ``Embed``, then `TypeError` is raised.
        
        file : `None`, `object`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of \
                ``Component``), Optional (Keyword only)
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        Raises
        ------
        TypeError
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``Embed`` nor as `list`, `tuple` of ``Embed``-s.
            - If `content` parameter was given as ``Embed``, meanwhile `embed` parameter was given as well.
            - If `message` was not given neither as ``Message``, `int`.
        ValueError
            If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Cannot editing interaction messages, which were created with `show_for_invoking_user_only=True`:
        
        ```
        DiscordException Not Found (404), code=10008: Unknown Message
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 5.: raise
        
        if isinstance(message, Message):
            message_id = message.id
        else:
            message_id = maybe_snowflake(message)
            if (message_id is None):
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got '
                    f'{message.__class__.__name__}, {message!r}.'
                )
        
        content, embed = validate_content_and_embed(content, embed, True)
        
        components = get_components_data(components, True)
        
        # Build payload
        message_data = {}
        
        # Discord docs say, content can be nullable, but nullable content is just ignored.
        if (content is not ...):
            message_data['content'] = content
        
        if (embed is not ...):
            if (embed is not None):
                embed = [embed.to_data() for embed in embed]
            
            message_data['embeds'] = embed
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not ...):
            message_data['components'] = components
        
        message_data = add_file_to_message_data(message_data, file, True, True)
        
        async with InteractionResponseContext(interaction_event, False, False):
            # We receive the new message data, but we do not update the message, so dispatch events can get the
            # difference.
            await self.http.interaction_followup_message_edit(
                application_id, interaction_event.id, interaction_event.token, message_id, message_data
            )
    
    
    async def interaction_followup_message_delete(self, interaction_event, message):
        """
        Deletes an interaction's followup message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        message : ``Message``, `int`
            The interaction followup's message to edit.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 5.: raise
        
        if isinstance(message, Message):
            message_id = message.id
        else:
            message_id = maybe_snowflake(message)
            if (message_id is None):
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got {message.__class__.__name__}; {message!r}.'
                )
        
        await self.http.interaction_followup_message_delete(
            application_id, interaction_event.id, interaction_event.token, message_id
        )
    
    
    async def interaction_followup_message_get(self, interaction_event, message_id):
        """
        Gets a previously sent message with an interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        message_id : `int`
            The webhook's message's identifier to get.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If `message_id` was not given as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        
        message_id_value = maybe_snowflake(message_id)
        if message_id_value is None:
            raise TypeError(
                f'`message_id` can be `int`, got {message_id.__class__.__name__}; {message_id!r}.'
            )
        
        message_data = await self.http.interaction_followup_message_get(
            application_id, interaction_event.id, interaction_event.token, message_id
        )
        
        return Message.from_data(message_data)



__all__ = ()

from ...webhook import WebhookType

from .shared import _convert_preinstanced, convert_icon, convert_nothing, convert_snowflake


def convert_webhook_type(name, data):
    return _convert_preinstanced('type', data, WebhookType)


WEBHOOK_CONVERTERS = {
    'application_id': convert_snowflake,
    'avatar_hash': convert_icon,
    'channel_id': convert_snowflake,
    'id': convert_snowflake,
    'name': convert_nothing,
    'type': convert_webhook_type,
}



__all__ = ('clear_default_plugin_variables',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.clear_default_variables)
def clear_default_plugin_variables():
    return PLUGIN_LOADER.clear_default_variables()



__all__ = ('ChannelMetadataGuildThreadPublic',)

from scarletio import copy_docs

from .fields import (
    parse_applied_tag_ids, parse_flags, put_applied_tag_ids_into, put_flags_into, validate_applied_tag_ids,
    validate_flags
)
from .flags import ChannelFlag

from .guild_thread_base import ChannelMetadataGuildThreadBase


class ChannelMetadataGuildThreadPublic(ChannelMetadataGuildThreadBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _created_at : `None`, `datetime`
        When the channel was created.
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    applied_tag_ids : `None`, `tuple` of `int`
         The tags' identifier which have been applied to the thread. Applicable for threads of a forum.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None`, `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    flags : ``ChannelFlag``
        The channel's flags.
    name : `str`
        The channel's name.
    open : `bool`
        Whether the thread channel is open.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    parent_id : `int`
        The channel's parent's identifier.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channels` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('applied_tag_ids', 'flags',)
    
    
    def __new__(
        cls,
        *,
        applied_tag_ids = ...,
        archived = ...,
        archived_at = ...,
        auto_archive_after = ...,
        flags = ...,
        created_at = ...,
        name = ...,
        parent_id = ...,
        open = ...,
        owner_id = ...,
        slowmode = ...,
    ):
        """
        Creates a new guild public thread channel metadata from the given parameters.
        
        Parameters
        ----------
        applied_tag_ids : `None`, `iterable` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread.
        archived : `bool`, Optional (Keyword only)
            Whether the thread is archived.
        archived_at : `None`, `datetime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        auto_archive_after : `int`, Optional (Keyword only)
            Duration in seconds to automatically archive the thread after recent activity.
        flags : ``ChannelFlag``, `int`, Optional (Keyword only)
            The channel's flags.
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the channel was created.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        open : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The channel's creator's identifier.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds what a user needs to wait between it's each message.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # applied_tag_ids
        if applied_tag_ids is ...:
            applied_tag_ids = None
        else:
            applied_tag_ids = validate_applied_tag_ids(applied_tag_ids)
        
        # flags
        if flags is ...:
            flags = ChannelFlag()
        else:
            flags = validate_flags(flags)
        
        # Construct
        self = ChannelMetadataGuildThreadBase.__new__(
            cls,
            archived = archived,
            archived_at = archived_at,
            auto_archive_after = auto_archive_after,
            created_at = created_at,
            name = name,
            open = open,
            owner_id = owner_id,
            parent_id = parent_id,
            slowmode = slowmode,
        )
        self.applied_tag_ids = applied_tag_ids
        self.flags = flags
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            applied_tag_ids = keyword_parameters.pop('applied_tag_ids', ...),
            archived = keyword_parameters.pop('archived', ...),
            archived_at = keyword_parameters.pop('archived_at', ...),
            auto_archive_after = keyword_parameters.pop('auto_archive_after', ...),
            flags = keyword_parameters.pop('flags', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            open = keyword_parameters.pop('open', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildThreadBase._is_equal_same_type(self, other):
            return False
        
        # applied_tag_ids
        if self.applied_tag_ids != other.applied_tag_ids:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildThreadBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildThreadPublic, cls)._create_empty()
        
        # applied_tag_ids
        self.applied_tag_ids = None
        
        # flags
        self.flags = ChannelFlag()
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.copy)
    def copy(self):
        new = ChannelMetadataGuildThreadBase.copy(self)
        
        applied_tag_ids = self.applied_tag_ids
        if (applied_tag_ids is not None):
            applied_tag_ids = (*applied_tag_ids,)
        new.applied_tag_ids = applied_tag_ids
        
        new.flags = self.flags
        
        return new
    
    
    def copy_with(
        self,
        *,
        applied_tag_ids = ...,
        archived = ...,
        archived_at = ...,
        auto_archive_after = ...,
        flags = ...,
        created_at = ...,
        name = ...,
        parent_id = ...,
        open = ...,
        owner_id = ...,
        slowmode = ...,
    ):
        """
        Copies the guild public thread channel metadata with the given fields.
        
        Parameters
        ----------
        applied_tag_ids : `None`, `iterable` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread.
        archived : `bool`, Optional (Keyword only)
            Whether the thread is archived.
        archived_at : `None`, `datetime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        auto_archive_after : `int`, Optional (Keyword only)
            Duration in seconds to automatically archive the thread after recent activity.
        flags : ``ChannelFlag``, `int`, Optional (Keyword only)
            The channel's flags.
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the channel was created.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        open : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The channel's creator's identifier.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds what a user needs to wait between it's each message.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # applied_tag_ids
        if applied_tag_ids is ...:
            applied_tag_ids = self.applied_tag_ids
            if (applied_tag_ids is not None):
                applied_tag_ids = (*applied_tag_ids,)
        else:
            applied_tag_ids = validate_applied_tag_ids(applied_tag_ids)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # Construct
        new = ChannelMetadataGuildThreadBase.copy_with(
            self,
            archived = archived,
            archived_at = archived_at,
            auto_archive_after = auto_archive_after,
            created_at = created_at,
            name = name,
            open = open,
            owner_id = owner_id,
            parent_id = parent_id,
            slowmode = slowmode,
        )
        new.applied_tag_ids = applied_tag_ids
        new.flags = flags
        return new
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            applied_tag_ids = keyword_parameters.pop('applied_tag_ids', ...),
            archived = keyword_parameters.pop('archived', ...),
            archived_at = keyword_parameters.pop('archived_at', ...),
            auto_archive_after = keyword_parameters.pop('auto_archive_after', ...),
            flags = keyword_parameters.pop('flags', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            open = keyword_parameters.pop('open', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildThreadBase._update_attributes(self, data)
        
        # applied_tag_ids
        self.applied_tag_ids = parse_applied_tag_ids(data)
        
        # flags
        self.flags = parse_flags(data)
    
    
    @copy_docs(ChannelMetadataGuildThreadBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildThreadBase._difference_update_attributes(self, data)
        
        # applied_tag_ids
        applied_tag_ids = parse_applied_tag_ids(data)
        if (self.applied_tag_ids != applied_tag_ids):
            old_attributes['applied_tag_ids'] = self.applied_tag_ids
            self.applied_tag_ids = applied_tag_ids
        
        # flags
        flags = parse_flags(data)
        if (self.flags != flags):
            flags = ChannelFlag(flags)
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildThreadBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildThreadBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # applied_tag_ids
        put_applied_tag_ids_into(self.applied_tag_ids, data, defaults)
        
        # flags
        put_flags_into(self.flags, data, defaults)
        
        return data



__all__ = ('create_partial_soundboard_sound_from_id', 'create_partial_soundboard_sound_from_partial_data')

from ...core import SOUNDBOARD_SOUNDS

from .fields import (
    parse_guild_id, parse_id, put_emoji_into, put_name_into, put_volume_into, validate_emoji, validate_name,
    validate_volume
)
from .soundboard_sound import SoundboardSound


SOUNDBOARD_SOUND_FIELD_CONVERTERS = {
    'emoji': (validate_emoji, put_emoji_into),
    'name': (validate_name, put_name_into),
    'volume': (validate_volume, put_volume_into),
}


def create_partial_soundboard_sound_from_id(sound_id, guild_id = 0):
    """
    Creates a partial soundboard sound from the given `sound_id`. If the sound already exists returns that instead.
    
    Parameters
    ----------
    sound_id : `int`
        The unique identifier number of the sound.
    guild_id : `int` = `0`, Optional
        The sound's guild's identifier.
    
    Returns
    -------
    sound : ``SoundboardSound``
    """
    try:
        sound = SOUNDBOARD_SOUNDS[sound_id]
    except KeyError:
        sound = SoundboardSound._create_empty(sound_id, guild_id)
        SOUNDBOARD_SOUNDS[sound_id] = sound
    
    return sound


def create_partial_soundboard_sound_from_partial_data(data):
    """
    Creates a soundboard instance from the given partial data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Soundboard data.
    
    Returns
    -------
    sound : ``SoundboardSound``
    """
    sound_id = parse_id(data)
    
    try:
        sound = SOUNDBOARD_SOUNDS[sound_id]
    except KeyError:
        sound = SoundboardSound._create_empty(sound_id, parse_guild_id(data))
        SOUNDBOARD_SOUNDS[sound_id] = sound
    
    return sound



__all__ = ()

import warnings

from scarletio import Compound

from ...guild import DiscoveryCategory, Guild, GuildDiscovery
from ...guild.discovery.utils import GUILD_DISCOVERY_FIELD_CONVERTERS
from ...http import DiscordHTTPClient, rate_limit_groups
from ...payload_building import build_edit_payload

from ..functionality_helpers import DiscoveryCategoryRequestCacher, DiscoveryTermRequestCacher
from ..request_helpers import get_guild_id


class ClientCompoundDiscoveryEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def guild_discovery_get(self, guild):
        """
        Requests and returns the guild's discovery metadata.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild what's discovery will be requested.
        
        Returns
        -------
        guild_discovery : ``GuildDiscovery``
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        guild_discovery_data = await self.http.guild_discovery_get(guild_id)
        return GuildDiscovery.from_data(guild_discovery_data)
    
    
    async def guild_discovery_edit(self, guild, discovery_template = None, **keyword_parameters):
        """
        Edits the guild's discovery metadata.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        
        > To modify sub-categories use the ``.guild_discovery_add_sub_category`` and the
        > ``.guild_discovery_delete_sub_category`` methods.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's discovery metadata will be edited or an existing discovery metadata object.
        
        discovery_template : `None`, ``GuildDiscovery`` = `None`, Optional
            Guild discovery to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to define which fields should be modified.
        
        Other Parameters
        ----------------
        emoji_discovery : `bool`, Optional (Keyword only)
            Whether guild info is shown when the respective guild's emojis are clicked.
        keywords : `None`, `iterable` of `str`, Optional (Keyword only)
            The set discovery search keywords for the guild.
        primary_category : ``DiscoveryCategory``, `int`, Optional (Keyword only)
            The primary discovery category of the guild.
        
        Returns
        -------
        guild_discovery : ``GuildDiscovery``
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        data = build_edit_payload(None, discovery_template, GUILD_DISCOVERY_FIELD_CONVERTERS, keyword_parameters)
        guild_discovery_data = await self.http.guild_discovery_edit(guild_id, data)
        return GuildDiscovery.from_data(guild_discovery_data)
    
    
    async def guild_discovery_add_subcategory(self, *positional_parameters, **keyword_parameters):
        """
        ``Client.guild_discovery_add_subcategory`` is deprecated and will be removed in 2023 Mar.
        Please use ``.guild_discovery_add_sub_category`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.guild_discovery_add_subcategory` is deprecated and will be removed in '
                f'2023 Mar. Please use `.guild_discovery_add_sub_category` instead.'
            ),
            FutureWarning
        )
        
        return await self.guild_discovery_add_sub_category(*positional_parameters, **keyword_parameters)
    
    
    async def guild_discovery_add_sub_category(self, guild, category):
        """
        Adds a discovery subcategory to the guild.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to what the discovery subcategory will be added.
        category : ``DiscoveryCategory``, `int`
            The discovery category or it's id what will be added as a subcategory.
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
            - If `category`is not ``DiscoveryCategory``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        A guild can have maximum `5` discovery subcategories.
        """
        guild_id = get_guild_id(guild)
        
        if isinstance(category, DiscoveryCategory):
            category_id = category.id
        elif isinstance(category, DiscoveryCategory.VALUE_TYPE):
            category_id = category
        else:
            raise TypeError(
                f'`category` can be `{DiscoveryCategory.__name__}`, `{DiscoveryCategory.VALUE_TYPE.__name__}`, '
                f'got {category.__class__.__name__}; {category!r}.'
            )
        
        await self.http.guild_discovery_add_sub_category(guild_id, category_id)
    
    
    async def guild_discovery_delete_subcategory(self, *positional_parameters, **keyword_parameters):
        """
        ``Client.guild_discovery_delete_subcategory`` is deprecated and will be removed in 2023 Mar.
        Please use ``.guild_discovery_delete_sub_category`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.guild_discovery_delete_subcategory` is deprecated and will be removed in '
                f'2023 Mar. Please use `.guild_discovery_delete_sub_category` instead.'
            ),
            FutureWarning
        )
        
        return await self.guild_discovery_delete_sub_category(*positional_parameters, **keyword_parameters)
    
    
    async def guild_discovery_delete_sub_category(self, guild, category):
        """
        Removes a discovery subcategory of the guild.
        
        The client must have `manage_guild` permission to execute this method.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to what the discovery subcategory will be removed from.
        category : ``DiscoveryCategory``, `int`
            The discovery category or it's id what will be removed from the subcategories.
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
            - If `category`is not ``DiscoveryCategory``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        A guild can have maximum `5` discovery subcategories.
        """
        guild_id = get_guild_id(guild)
        
        if isinstance(category, DiscoveryCategory):
            category_id = category.id
        elif isinstance(category, DiscoveryCategory.VALUE_TYPE):
            category_id = category
        else:
            raise TypeError(
                f'`category` can be `{DiscoveryCategory.__name__}`, `{DiscoveryCategory.VALUE_TYPE.__name__}`, '
                f'got {category.__class__.__name__}; {category!r}.'
            )
        
        await self.http.guild_discovery_delete_sub_category(guild_id, category_id)
    
    
    async def discovery_category_get_all(self):
        """
        Returns a list of discovery categories, which can be used when editing guild discovery.
        
        This method is a coroutine.
        
        Returns
        -------
        discovery_categories : `list` of ``DiscoveryCategory``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        discovery_category_datas = await self.http.discovery_category_get_all(None)
        return [
            DiscoveryCategory.from_data(discovery_category_data)
            for discovery_category_data in discovery_category_datas
        ]
    
    
    # Add cached, so even tho the first request fails with `ConnectionError` will not be raised.
    discovery_category_get_all = DiscoveryCategoryRequestCacher(
        discovery_category_get_all,
        3600.0,
        [*DiscoveryCategory.INSTANCES.values()],
    )
    
    
    async def discovery_validate_term(self, term):
        """
        Checks whether the given discovery search term is valid.
        
        This method is a coroutine.
        
        Parameters
        ----------
        term : `str`
            Discovery term to validate.
        
        Returns
        -------
        valid : `bool`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.http.discovery_validate_term({'term': term})
        return data['valid']
    
    
    discovery_validate_term = DiscoveryTermRequestCacher(
        discovery_validate_term,
        86400.0,
        rate_limit_groups.discovery_validate_term,
    )



__all__ = ('ClientUserPBase',)

from itertools import count

from scarletio import copy_docs

from ...activity import Activity, ActivityType

from ..activity_change import ActivityChange
from ..activity_update import ActivityUpdate

from .client_user_base import ClientUserBase
from .fields import (
    parse_activities, parse_status, parse_statuses, validate_activities, validate_status, validate_statuses
)
from .flags import UserFlag
from .preinstanced import Status


ACTIVITY_TYPE_CUSTOM = ActivityType.custom


class ClientUserPBase(ClientUserBase):
    """
    Base class for discord users and clients. This class is used as ``User`` superclass only if presence is enabled,
    so by default.
    
    Attributes
    ----------
    activities : `None`, `list` of ``Activity``
        A list of the client's activities. Defaults to `None`
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The user's avatar's type.
    avatar_decoration_hash : `int`
        The user's avatar decoration's hash in `uint128`.
    avatar_decoration_type : ``IconType``
        The user's avatar decoration's type.
    banner_color : `None`, ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    bot : `bool`
        Whether the user is a bot or a user account.
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    display_name : `None`, `str`
        The user's non-unique display name.
    flags : ``UserFlag``
        The user's user flags.
    guild_profiles : `dict` of (`int`, ``GuildProfile``) items
        A dictionary, which contains the user's guild profiles. If a user is member of a guild, then it should
        have a respective guild profile accordingly.
    id : `int`
        The user's unique identifier number.
    name : str
        The user's name.
    status : `Status`
        The user's display status.
    statuses : `None`, `dict` of (`str`, `str`) items
        The user's statuses for each platform.
    thread_profiles : `None`, `dict` (``Channel``, ``ThreadProfile``) items
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
    """
    __slots__ = ('activities', 'status', 'statuses')
    
    def __new__(
        cls,
        *,
        activities = ...,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        bot = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
        status = ...,
        statuses = ...,
    ):
        """
        Creates a new partial user with the given fields.
        
        Parameters
        ----------
        activities : `iterable`, `list` of ``Activity``, Optional (Keyword only)
            A list of the client's activities.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        avatar_decoration : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar decoration.
        banner : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's banner.
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        bot : `bool`, Optional (Keyword only)
            Whether the user is a bot or a user account.
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        display_name : `None`, `str`, Optional (Keyword only)
            The user's non-unique display name.
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        name : `str`, Optional (Keyword only)
            The user's name.
        status : `Status`, `str`, Optional (Keyword only)
            The user's display status.
        statuses : `None`, `dict` of (`str`, `str`) items, Optional (Keyword only)
            The user's statuses for each platform.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # activities
        if activities is ...:
            activities = None
        else:
            activities = validate_activities(activities)
        
        # status
        if status is ...:
            status = Status.offline
        else:
            status = validate_status(status)
        
        # statuses
        if statuses is ...:
            statuses = None
        else:
            statuses = validate_statuses(statuses)
        
        self = ClientUserBase.__new__(
            cls,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            bot = bot,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
        )
        
        self.activities = activities
        self.status = status
        self.statuses = statuses
        return self
        
    
    @classmethod
    @copy_docs(ClientUserBase._from_client)
    def _from_client(cls, client, include_internals):
        self = super(ClientUserPBase, cls)._from_client(client, include_internals)
        
        # activities
        activities = client.activities
        if (activities is not None):
            activities = [activity.copy() for activity in activities]
        self.activities = activities
        
        # status
        self.status = client.status
        
        # statuses
        statuses = client.statuses
        if (statuses is not None):
            statuses = statuses.copy()
        self.statuses = statuses
        
        return self
    
    
    @copy_docs(ClientUserBase._set_default_attributes)
    def _set_default_attributes(self):
        ClientUserBase._set_default_attributes(self)
        
        self.status = Status.offline
        self.statuses = None
        self.activities = None
    
    
    @copy_docs(ClientUserBase._update_presence)
    def _update_presence(self, data):
        self.activities = parse_activities(data)
        self.status = parse_status(data)
        self.statuses = parse_statuses(data)
    
    
    @copy_docs(ClientUserBase._difference_update_presence)
    def _difference_update_presence(self, data):
        old_attributes = {}
        
        statuses = parse_statuses(data)
        if self.statuses != statuses:
            old_attributes['statuses'] = self.statuses
            self.statuses = statuses
            
        status = parse_status(data)
        if self.status is not status:
            old_attributes['status'] = self.status
            self.status = status
        
        activity_datas = data['activities']
        
        old_activities = self.activities
        new_activities = None
        
        if activity_datas:
            if old_activities is None:
                for activity_data in activity_datas:
                    activity = Activity.from_data(activity_data)
                    
                    if new_activities is None:
                        new_activities = []
                    
                    new_activities.append(activity)
                
                activity_change = ActivityChange.from_fields(new_activities, None, None)
                
            else:
                added_activities = None
                updated_activities = None
                removed_activities = old_activities.copy()
                
                for activity_data in activity_datas:
                    activity_type = activity_data['type']
                    for index in range(len(removed_activities)):
                        activity = removed_activities[index]
                        
                        if activity_type != activity.type:
                            continue
                            
                        if activity_data['id'] != activity.discord_side_id:
                            continue
                        
                        del removed_activities[index]
                        
                        activity_old_attributes = activity._difference_update_attributes(activity_data)
                        if activity_old_attributes:
                            activity_update = ActivityUpdate.from_fields(activity, activity_old_attributes)
                            
                            if updated_activities is None:
                                updated_activities = []
                            
                            updated_activities.append(activity_update)
                        
                        if new_activities is None:
                            new_activities = []
                        
                        new_activities.append(activity)
                        break
                    else:
                        activity = Activity.from_data(activity_data)
                        
                        if new_activities is None:
                            new_activities = []
                        
                        new_activities.append(activity)
                        
                        if added_activities is None:
                            added_activities = []
                        
                        added_activities.append(activity)
                
                if not removed_activities:
                    removed_activities = None
                
                if None is added_activities is updated_activities is removed_activities:
                    activity_change = None
                else:
                    activity_change = ActivityChange.from_fields(
                        added_activities, updated_activities, removed_activities
                    )
        
        else:
            if old_activities is None:
                activity_change = None
            else:
                activity_change = ActivityChange.from_fields(None, None, old_activities)
        
        if (activity_change is not None):
            old_attributes['activities'] = activity_change
        
        self.activities = new_activities
        
        return old_attributes
    
    
    @copy_docs(ClientUserBase.copy)
    def copy(self):
        new = ClientUserBase.copy(self)
        
        # activities
        activities = self.activities
        if (activities is not None):
            activities = [activity.copy() for activity in activities]
        new.activities = activities
        
        # status
        new.status = self.status
        
        # statuses
        statuses = self.statuses
        if (statuses is not None):
            statuses = statuses.copy()
        new.statuses = statuses
        
        return new
    
    
    def copy_with(
        self,
        *,
        activities = ...,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        bot = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
        status = ...,
        statuses = ...,
    ):
        """
        Copies the user with the given fields.
        
        Parameters
        ----------
        activities : `iterable`, `list` of ``Activity``, Optional (Keyword only)
            A list of the client's activities.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        avatar_decoration : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar decoration.
        banner : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's banner.
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        bot : `bool`, Optional (Keyword only)
            Whether the user is a bot or a user account.
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        display_name : `None`, `str`, Optional (Keyword only)
            The user's non-unique display name.
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        name : `str`, Optional (Keyword only)
            The user's name.
        status : `Status`, `str`, Optional (Keyword only)
            The user's display status.
        statuses : `None`, `dict` of (`str`, `str`) items, Optional (Keyword only)
            The user's statuses for each platform.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # activities
        if activities is ...:
            activities = self.activities
            if (activities is not None):
                activities = [activity.copy() for activity in activities]
        else:
            activities = validate_activities(activities)
        
        # status
        if status is ...:
            status = self.status
        else:
            status = validate_status(status)
        
        # statuses
        if statuses is ...:
            statuses = self.statuses
            if (statuses is not None):
                statuses = statuses.copy()
        else:
            statuses = validate_statuses(statuses)
        
        new = ClientUserBase.copy_with(
            self,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            bot = bot,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
        )
        
        new.activities = activities
        new.status = status
        new.statuses = statuses
        return new
    
    
    @copy_docs(ClientUserBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = ClientUserBase._get_hash_partial(self)
        
        # activities
        activities = self.activities
        if (activities is not None):
            hash_value ^= len(activities) << 2
            
            for mask_shift, activity in zip(count(4, 4), activities, 3):
                hash_value ^= hash(activity) | (0xf << mask_shift)
        
        # status
        hash_value ^= hash(self.status)
        
        # statuses
        hash_value ^= hash(tuple(self.statsues.items()))
        
        return hash_value

    
    @property
    @copy_docs(ClientUserBase.activity)
    def activity(self):
        activities = self.activities
        if activities is None:
            activity = None
        else:
            for activity in activities:
                if activity.type is not ACTIVITY_TYPE_CUSTOM:
                    break
            else:
                activity = None
        
        return activity
    
    
    @property
    @copy_docs(ClientUserBase.custom_activity)
    def custom_activity(self):
        activities = self.activities
        if activities is None:
            activity = None
        else:
            for activity in activities:
                if activity.type is ACTIVITY_TYPE_CUSTOM:
                    break
            else:
                activity = None
        
        return activity
    
    
    @property
    @copy_docs(ClientUserBase.platform)
    def platform(self):
        statuses = self.statuses
        if (statuses is not None):
            actual_status_value = self.status.value
            for platform, status_value in statuses.items():
                if actual_status_value == status_value:
                    return platform
        
        return ''
    
    
    @copy_docs(ClientUserBase.get_status_by_platform)
    def get_status_by_platform(self, platform):
        statuses = self.statuses
        if (statuses is not None):
            try:
                status_value = statuses[platform]
            except KeyError:
                pass
            else:
                return Status.get(status_value)
        
        return Status.offline



__all__ = ('add_external_command_route', 'get_external_command_routes',)

import sys
from os.path import exists, join as join_paths, isdir as is_directory, isfile as is_file

from .constants import EXTERNAL_IMPORT_ROUTES_FILE, PYTHON_FILE_POSTFIX_NAMES


def _validate_import_route(import_route):
    """
    Returns whether the given import route refers to an existing file / directory.
    
    Parameters
    ----------
    import_route : `str`
        The import route to check out.
    
    Returns
    -------
    exists : `bool`
    """
    import_route_split = import_route.split('.')
    
    for path in sys.path:
        full_path = join_paths(path, *import_route_split)
        if exists(full_path):
            if is_directory(full_path):
                if is_file(join_paths(full_path, '__init__.py')):
                    return True
        
        else:
            for postfix in PYTHON_FILE_POSTFIX_NAMES:
                if is_file(full_path + postfix):
                    return True
    
    return False


def _save_external_command_routes(command_routes):
    """
    Saves the import routes.
    
    Parameters
    ----------
    command_routes : `set` of `str`
    """
    content = '\n'.join(sorted(command_routes))
    
    with open(EXTERNAL_IMPORT_ROUTES_FILE, 'w') as file:
        file.write(content)


def get_external_command_routes(*, fix_errors=True):
    """
    Reads the external command routes.
    
    Parameters
    ----------
    fix_errors : `bool` = `True`, Optional (Keyword only)
        Whether read errors should be fixed.
    """
    command_routes = set()
    
    if not exists(EXTERNAL_IMPORT_ROUTES_FILE):
        return command_routes
    
    
    with open(EXTERNAL_IMPORT_ROUTES_FILE, 'r') as file:
        content = file.read()
    
    error_found = False
    
    for import_route in content.splitlines():
        import_route = import_route.strip()
        
        if _validate_import_route(import_route):
            command_routes.add(import_route)
        
        else:
            error_found = True
    
    if error_found and fix_errors:
        _save_external_command_routes(command_routes)
    
    return command_routes


def add_external_command_route(command_route, *, validate_command_route=False):
    """
    Adds a new external command route.
    
    Parameters
    ----------
    command_route : `str`
        The command route to add.
    validate_command_route : `bool` = `False`, Optional (Keyword only)
        Whether the existence of the added command route should be checked initially.
    
    Returns
    -------
    command_route_added : `bool`
        Whether the command route was added.
    """
    if validate_command_route:
        if not _validate_import_route(command_route):
            return False
    
    command_routes = get_external_command_routes(fix_errors=False)
    command_routes.add(command_route)
    _save_external_command_routes(command_routes)
    return True



__all__ = ('GuildJoinRequestStatus',)

from ...bases import Preinstance as P, PreinstancedBase


class GuildJoinRequestStatus(PreinstancedBase):
    """
    Represents the status of a ``GuildJoinRequest``.
    
    Attributes
    ----------
    name : `str`
        The name of the guild join request status.
    value : `str`
        The Discord side identifier value of the guild join request status.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``GuildJoinRequestStatus``) items
        Stores the predefined ``GuildJoinRequestStatus``-s.
    VALUE_TYPE : `type` = `str`
        The guild join request statuses' values' type.
    DEFAULT_NAME : `str` = `''`
        The default name of the guild join request statuses. Guild join request statuses have their name generated from
        their value, so at their case it is not applicable.
    
    Every predefined guild join request status can be accessed as class attribute as well:
    
    +-----------------------+-----------+-----------+
    | Class attribute names | Name      | Value     |
    +=======================+===========+===========+
    | approved              | approved  | APPROVED  |
    +-----------------------+-----------+-----------+
    | pending               | pending   | PENDING   |
    +-----------------------+-----------+-----------+
    | rejected              | rejected  | REJECTED  |
    +-----------------------+-----------+-----------+
    | started               | started   | STARTED   |
    +-----------------------+-----------+-----------+
    """
    
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    
    __slots__ = ()
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new guild join request status from the given value.
        
        Parameters
        ----------
        value : `str`
            The guild join request status's identifier value.
        
        Returns
        -------
        self : ``GuildJoinRequestStatus``
            The guild join request status.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value.lower().replace('_', ' ')
        self.INSTANCES[value] = self
        return self
    
    def __repr__(self):
        """Returns the representation of the guild join request status."""
        return f'{self.__class__.__name__}(value = {self.value!r})'
    
    approved = P('APPROVED', 'approved')
    pending = P('PENDING', 'pending')
    rejected = P('REJECTED', 'rejected')
    started = P('STARTED', 'started')




__all__ = ()

from .constants import (
    EMBEDDED_ACTIVITY_UPDATE_CREATE, EMBEDDED_ACTIVITY_UPDATE_DELETE, EMBEDDED_ACTIVITY_UPDATE_UPDATE,
    EMBEDDED_ACTIVITY_UPDATE_USER_ADD, EMBEDDED_ACTIVITY_UPDATE_USER_DELETE
)
from .embedded_activity_state import EmbeddedActivityState


def difference_handle_embedded_activity_update_event(data):
    """
    Handles embedded activity events returning the embedded activity state and the changes with it.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Embedded activity update event.
    
    Returns
    -------
    embedded_activity_state : ``EmbeddedActivityState``
        The updated embedded activity state.
    
    changes : `list` of `tuple` (`int`, `object`)
        Change entries. Each tuple` has `2` elements:
        
        - The action identifier of the change:
        
            Can be one of the following:
            
            +---------------------------------------+-------+
            | Respective name                       | Value |
            +=======================================+=======+
            | EMBEDDED_ACTIVITY_UPDATE_NONE         | 0     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_CREATE       | 1     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_DELETE       | 2     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_UPDATE       | 3     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_ADD     | 4     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_DELETE  | 5     |
            +---------------------------------------+-------+
        
        - The value in the context of the action.
            
            If `action` is `EMBEDDED_ACTIVITY_UPDATE_UPDATE`, this value will contain the updated attributes of the
            activity.
            
            +-------------------+-----------------------------------+
            | Keys              | Values                            |
            +===================+===================================+
            | assets            | `None`, ``ActivityAssets``        |
            +-------------------+-----------------------------------+
            | created_at        | `datetime`                        |
            +-------------------+-----------------------------------+
            | details           | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | flags             | ``ActivityFlag``                  |
            +-------------------+-----------------------------------+
            | name              | `str`                             |
            +-------------------+-----------------------------------+
            | metadata          | ``ActivityMetadataBase``          |
            +-------------------+-----------------------------------+
            | party             | `None`, ``ActivityParty``         |
            +-------------------+-----------------------------------+
            | secrets           | `None`, ``ActivitySecrets``       |
            +-------------------+-----------------------------------+
            | session_id        | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | state             | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | sync_id           | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | timestamps        | `None`, `ActivityTimestamps``     |
            +-------------------+-----------------------------------+
            | url               | `None`, `str`                     |
            +-------------------+-----------------------------------+
            
            If `action` is `EMBEDDED_ACTIVITY_UPDATE_USER_ADD`, `EMBEDDED_ACTIVITY_UPDATE_USER_DELETE`, it will
            contain the joined or left user's identifier.
    """
    embedded_activity_state, is_created = EmbeddedActivityState.from_data_is_created(data)
    
    changes = []
    
    # If the object was just created, we have 2 cases:
    # Just came to cache as a deleted activity / it was created
    if is_created:
        if embedded_activity_state.user_ids:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_CREATE, None))
        else:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    else:
        joined_user_ids, left_user_ids = embedded_activity_state._difference_update_user_ids(data)
        for user_id in joined_user_ids:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_ADD, user_id))
        
        for user_id in left_user_ids:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_DELETE, user_id))
        
        if embedded_activity_state.user_ids:
            activity_old_attributes = embedded_activity_state._difference_update_activity(data)
            if activity_old_attributes:
                changes.append((EMBEDDED_ACTIVITY_UPDATE_UPDATE, activity_old_attributes))
        else:
            embedded_activity_state._update_activity(data)
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    return embedded_activity_state, changes


def handle_embedded_activity_update_event(data):
    """
    Handles an embedded activity event.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Embedded activity update event.
    """
    embedded_activity_state, is_created = EmbeddedActivityState.from_data_is_created(data)
    if not is_created:
        embedded_activity_state._update_user_ids(data)
        embedded_activity_state._update_activity(data)
